/*
 * debug.c - Debugger interface
 *
 * Written 2003 by Werner Almesberger, Caltech Netlab FAST project
 */

#include <inttypes.h>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#include "arch.h"
#include "dwarfx.h"
#include "ptrace.h"

#include "util.h"
#include "error.h"
#include "data.h"
#include "op.h"
#include "process.h"
#include "sbug.h"


/* ----- Pointer and array conversions ------------------------------------- */


static const struct dwx_loc_expr *frame_base(const struct process *process)
{
    const struct dwx_obj *obj;
    arch_addr_t pc;

    pc = process->proc->arch_ops->get_pc(process->proc->arch_dsc);
    obj = dwx_find_fn_by_addr(process->objs,pc);
    return obj ? dwx_loc_expr_here(obj->u.fn.frame_addr,pc) : NULL;
}


static arch_addr_t var_data_location(const struct process *process,
  const struct dwx_loc_expr *loc)
{
    struct dwx_loc here;

    dwx_loc_setup_hack(process->proc->arch_ops,process->proc->arch_dsc,
      frame_base(process));
    here = dwx_loc_here(loc,
      process->proc->arch_ops->get_pc(process->proc->arch_dsc));
    assert(here.reg == -1);
    return here.addr;
             
}


/* ----- Operators: & ------------------------------------------------------ */


static struct value *do_op_address(struct value *value)
{
    struct value *new;
    const struct dwx_type *type,*stripped;

    if (value && value->type == &lvalue_type)
	value = dereference_lvalue(value);
    if (!value || private_type(value->type))
	clerrorf("invalid operand (variable object expected)");
    type = value->type;
    stripped = dwx_strip_type(type);
    if (stripped && stripped->kind == dwx_array) type = stripped->u.array.type;
    new = new_value(dwx_add_ptr(type));
    new->flags |= VF_MALLOCED;
    new->process = get_process(value->process);
    if (value->flags & VF_REMOTE) {
	new->u.value = alloc_t(arch_addr_t);
	*(arch_addr_t *) new->u.value = var_data_location(value->process,
	  value->u.addr.loc)+value->u.addr.offset;
    }
    else {
	if (value->process) {
	    clerrorf("invalid operand for &");
#if 0
	    /*
	     * We can't do this, because we'd lose the process context, so the
	     * whole thing could never be dereferenced.
	     * (FFS: consider chaining through value->owner)
	     */
	    new->u.value = alloc_t(arch_addr_t);
	    *(arch_addr_t *) new->u.value =
	      *(arch_addr_t *) value->u.value;
#endif
	}
	else {
	    share(new,value);
	    new->u.value = alloc_t(void *);
	    *(void **) new->u.value = value->u.value;
	}
    }
    put(value);
    return new;
}


struct value *op_address(const struct op *self)
{
    return do_op_address(evaluate_lvalue(self->u.arg.a));
}


/* ----- Operators: * ------------------------------------------------------ */


static uint8_t zero_addr_expr[] = { DW_OP_lit0 };

static struct dwx_loc_expr zero_addr = {
    .start = 0,
    .end = ~0U,	/* @@@ ugly and actually wrong */
    .expr = zero_addr_expr,
    .length = sizeof(zero_addr_expr),
    .next = NULL,
};


static struct value *dereference_array(struct value *value,
  const struct dwx_type *type)
{
    struct value *res;

    res = new_value(type);
    res->u = value->u;
    if (value->flags & VF_REMOTE) {
	res->flags |= VF_REMOTE;
	res->process = get_process(value->process);
    }
    else share(res,value);
    put(value);
    return res;
}


/* @@@ NEEDS CLEANUP */
struct value *op_dereference(const struct op *self)
{
    struct value *value,*res;
    const struct dwx_type *type;

    value = evaluate_lvalue(self->u.arg.a);
    if (value && value->type == &lvalue_type)
	value = dereference_lvalue(value);
    if (!value || private_type(value->type))
	clerrorf("invalid operand type (data object expected)");
    type = dwx_strip_type(value->type);
    if (type->kind == dwx_array)
	return dereference_array(value,type->u.array.type);
    if (type->kind != dwx_ptr)
	clerrorf("invalid operand type (pointer expected)");

    type = type->u.modifier.type;
    if (!type) clerrorf("attempting to dereference void pointer");
    res = new_value(type);
    res->process = get_process(value->process);

    res->flags |= VF_REMOTE;
    res->u.addr.loc = &zero_addr;

    /* Pointer in remote process */
    if (value->flags & VF_REMOTE) {
	/*
	 * Do not express as pointer, but as data reference, so that we can use
	 * it as lvalue in assignments.
	 */
	check_process_termination(res->process);
	pt_read(res->process->proc,
	  var_data_location(value->process,value->u.addr.loc)+
	  value->u.addr.offset,
	  &res->u.addr.offset,sizeof(arch_addr_t));
	put(value);
	return res;
    }

    /*
     * Pointer to remote process.
     * Express as reference.
     */
    if (value->process) {
	res->u.addr.offset = *(arch_addr_t *) value->u.value;
	put(value);
	return res;
    }

    res->flags &= ~VF_REMOTE;

    /* Locally generated pointer */
    if (value->owner) {
	res->owner = get(value->owner);
	res->u.value = *(void **) value->u.value;
	put(value);
	return res;
    }

    /*
     * Pointer to local process, but not locally generated pointer.
     * This case has "can't happen" written all over it, but we still handle
     * it, just in case.
     */
    if (!value->process) {
	res->u.value = *(void **) value->u.value;
	put(value);
	return res;
    }

    /* Change this when done re-arranging cases @@@ */
    abort();
}


/* ----- Operators: -> ----------------------------------------------------- */


struct value *op_ptr(const struct op *self)
{
    /*
     * Not static, since we may recurse
     */
    struct op deref = {
	.fn = &op_dereference
    };
    struct op member = {
	.fn = &op_member,
    };

    deref.u.arg.a = self->u.arg.a;
    member.u.arg.a = &deref;
    member.u.arg.b = self->u.arg.b;
    return op_member(&member);
}


/* ----- Operators: + extension -------------------------------------------- */


static struct value *make_rvalue(struct value *value)
{
    struct value *new;
    int size;

    if (!value) return value;
    if (value->type == &lvalue_type) value = dereference_lvalue(value);
    if (!value) return value;
    if (private_type(value->type)) return value;
    if (!(value->flags & VF_REMOTE)) return value;
    if (dwx_strip_type(value->type)->kind == dwx_array)
	return do_op_address(value);
    assert(!private_type(value->type));
    new = new_value(value->type);
    new->flags |= VF_MALLOCED;
    new->process = get_process(value->process);
    size = dwx_sizeof(value->type);
    new->u.value = alloc(size);
    check_process_termination(value->process);
    pt_read(value->process->proc,
      var_data_location(value->process,value->u.addr.loc)+value->u.addr.offset,
      new->u.value,size);
    put(value);
    return new;
}


static struct value *advance_pointer(struct value *ptr,int ndx)
{
    struct value *res;

    assert(ptr->flags & VF_MALLOCED);
    res = new_value(ptr->type);
    res->flags = ptr->flags;
    res->process = get_process(ptr->process);
    res->owner = get(ptr->owner);
    if (ptr->type->u.modifier.type) /* sizeof(*(void *)) is 1 */
	ndx *= dwx_sizeof(ptr->type->u.modifier.type);
    if (ptr->process) {
	res->u.value = alloc_t(arch_addr_t);
	*(arch_addr_t *) res->u.value = *(arch_addr_t *) ptr->u.value+ndx;
    }
    else {
	res->u.value = alloc_t(void *);
	*(void **) res->u.value = *(void **) ptr->u.value+ndx;
    }
    put(ptr);
    return res;
}


int is_pointer(struct value **value)
{
    struct op addr = {
	.fn = op_address
    };
    const struct dwx_type *type;

    if (!*value || private_type((*value)->type)) return 0;
    type = dwx_strip_type((*value)->type);
    if (type->kind == dwx_ptr) return 1;
    if (type->kind != dwx_array) return 0;
    addr.u.arg.a = op_value(*value);
    *value = op_address(&addr);
    destroy_expr(addr.u.arg.a);
    return 1;
}


int sbug_op_add(struct value *a,struct value *b,struct value **res)
{
    struct value *ptr;
    int ndx;

    if (is_pointer(&a)) {
	ptr = make_rvalue(a);
	ndx = value_int(b);
    }
    else if (is_pointer(&b)) {
	ptr = make_rvalue(b);
	ndx = value_int(a);
    }
    else return 0;
    *res = advance_pointer(ptr,ndx);
    return 1;
}


/* ----- Operators: - extension -------------------------------------------- */


static struct value *ptr_diff(struct value *a,struct value *b)
{
    struct value *res;
    const struct dwx_type *type;
    unsigned size;

    a = make_rvalue(a);
    b = make_rvalue(b);
    if (!dwx_type_equal(a->type,b->type)) clerrorf("incompatible types");
    if (a->process != b->process) clerrorf("incompatible process contexts");
    type = a->type->u.modifier.type;
    size = type ? dwx_sizeof(type) : 1;
    res = int_value((a->process ?
      *(void **) a->u.value-*(void **) b->u.value :
      *(arch_addr_t *) a->u.value-*(arch_addr_t *) b->u.value
      )/size);
    put(a);
    put(b);
    return res;
}


int sbug_op_sub(struct value *a,struct value *b,struct value **res)
{
    if (is_pointer(&a) && is_pointer(&b)) {
	*res = ptr_diff(a,b);
	return 1;
    }
    if (!is_pointer(&a)) {
	if (is_pointer(&b))
	    clerrorf("cannot subtract pointer from non-pointer");
	return 0;
    }
    *res = advance_pointer(make_rvalue(a),-value_int(b));
    return 1;
}


/* ----- Operators: assignment extension (rvalue) -------------------------- */


static void write_reg(const struct pt_process *proc,int reg,const void *data,
  unsigned size)
{
    arch_reg_t value;

    assert(sizeof(value) >= size);
    switch (size) {
	case 1:
	    value = *(const uint8_t *) data;
	    break;
	case 2:
	    value = *(const uint16_t *) data;
	    break;
	case 4:
	    value = *(const uint32_t *) data;
	    break;
	case 8:
	    value = *(const uint64_t *) data;
	    break;
	default:
	    abort();
    }
    proc->arch_ops->set_reg(proc->arch_dsc,reg,value);
}


static void assign_var(const struct value *lvalue,const void *data,int size)
{
    const struct pt_process *proc = lvalue->process->proc;
    unsigned long addr = proc->arch_ops->get_pc(proc->arch_dsc);
    const struct dwx_loc_expr *walk;
    int found = 0;

    dwx_loc_setup_hack(proc->arch_ops,proc->arch_dsc,
      frame_base(lvalue->process));
    for (walk = lvalue->u.addr.loc; walk; walk = walk->next)
	if (walk->start <= addr && walk->end > addr) {
	    struct dwx_loc loc;

	    found = 1;
	    loc = dwx_loc_1(walk);
	    if (loc.reg == -1)
		pt_write(lvalue->process->proc,lvalue->u.addr.offset+loc.addr,
		  data,size);
	    else {
		assert(!lvalue->u.addr.offset); /* @@@ well, not quite ... */
		write_reg(proc,loc.reg,data,size);
	    }
	}
    if (!found) fprintf(stderr,"location unknown for address 0x%lx\n",addr);
}


int sbug_assign_hook(struct value *lvalue,struct value **rvalue)
{
    int size;

    *rvalue = make_rvalue(*rvalue);
    if (!lvalue || private_type(lvalue->type) || !(lvalue->flags & VF_REMOTE))
	return 0;
    if (private_type((*rvalue)->type)) clerrorf("invalid type in rvalue");
    if (!dwx_type_compatible((*rvalue)->type,lvalue->type))
	clerrorf("incompatible types in assignment");
    if ((*rvalue)->process != lvalue->process &&
      dwx_contains_pointers(lvalue->type))
	clerrorf("incompatible process contexts");
    size = dwx_sizeof((*rvalue)->type);
    check_process_termination(lvalue->process);
    /*
     * dwarfx should provide this
     */
    assign_var(lvalue,(*rvalue)->u.value,size);
    put(lvalue);
    return 1;
}


/* ----- Data access ------------------------------------------------------- */


struct value *dereference_object(struct value *value)
{
    return make_rvalue(value);
}
