/*
 * id.c - Identifier lookup
 *
 * Written 2003 by Werner Almesberger, Caltech Netlab FAST project
 */

#include <stdarg.h>
#include <stdlib.h>
#include <assert.h>

#include "arch.h"
#include "dwarfx.h"

#include "util.h"
#include "error.h"
#include "data.h"
#include "op.h"
#include "process.h"
#include "id.h"


struct dwx_type id_type;


/* ----- Identifier evaluation --------------------------------------------- */


static struct value *obj_to_value(struct process *process,
  const struct dwx_obj *obj)
{
    struct value *value;

    switch (obj->kind) {
	case dwx_obj_fn:
	case dwx_obj_label:
	    value = new_value(&location_type);
	    value->u.obj = obj;
	    break;
	case dwx_obj_var:
	    value = new_value(obj->u.var.type);
	    value->flags |= VF_REMOTE;
	    value->u.addr.loc = obj->u.var.loc;
	    value->u.addr.offset = 0;
	    break;
	default:
	    abort();
    }
    value->process = process;
    return value;
}


static const struct dwx_obj *try_local(const struct process *process,
  const char *name)
{
    const struct dwx_obj *local;

    if (!process->current_breakpoint)
	return dwx_find_item(process->objs->children,name);
    local = process->current_breakpoint->obj;
    if (local->kind == dwx_obj_label) local = local->parent;
    assert(local->kind == dwx_obj_fn);
    return dwx_find_item(local->children,name);
}


static const struct dwx_type *try_local_type(const struct process *process,
  enum dwx_kind kind,const char *name)
{
    const struct dwx_obj *local;

    if (!process->current_breakpoint)
	return dwx_find_type(process->objs->children,kind,name);
    local = process->current_breakpoint->obj;
    if (local->kind == dwx_obj_label) local = local->parent;
    assert(local->kind == dwx_obj_fn);
    return dwx_find_type(local->children,kind,name);
}


struct value *evaluate_id_left(const struct op *expr)
{
    struct value *value;

    if (!expr) clerrorf("null expression");
    value = expr->fn(expr);
    if (value && value->type == &lvalue_type)
	value = dereference_lvalue(value);
    if (!value) clerrorf("undefined value as identifier");
    return value;
}


struct value *evaluate_id(struct value *value,...)
{
    struct value *res;
    va_list ap;

    va_start(ap,value);
    while (1) {
	int goal;

	goal = va_arg(ap,int);
	if (!goal) {
	    if (value->type == &id_type)
		clerrorf("unknown or incompatible identifier \"%s\"",
		  value->u.id.name);
	    else clerrorf("unknown or incompatible identifier");
	}
	switch (goal) {
	    const struct dwx_type *type;
	    const struct dwx_obj *obj;

	    case dwx_obj_fn:
	    case dwx_obj_label:
		if (value->type == &location_type) {
		    if (value->u.obj->kind != goal) continue;
		    res = get(value);
		    break;
		}
		/* fall through */
	    case dwx_obj_var:
		if (value->type != &id_type) continue;
		if (!value->u.id.base) {
		    if (!value->process) continue;
		    obj = try_local(value->process,value->u.id.name);
		}
		else {
		    obj = dwx_find_item(value->u.id.base->children,
		       value->u.id.name);
		}
		if (!obj || obj->kind != goal) continue;
		res = obj_to_value(get_process(value->process),obj);
		break;
	    case dwx_enum:
	    case dwx_struct:
	    case dwx_union:
	    case dwx_typedef:
		if (value->type != &id_type) continue;
		if (!value->u.id.base) {
		    if (value->process) 
			type = try_local_type(value->process,goal,
			  value->u.id.name);
		    else {
			if (sbug) continue;
			type = dwx_lookup_type(goal,value->u.id.name,-1,0);
		    }
		}
		else {
		    type = dwx_find_type(value->u.id.base->children,
		       goal,value->u.id.name);
		}
		if (!type) continue;
		res = type_value(type);
		/*
		 * Record process context for now. Might be useful for some
		 * future compatibility checks, e.g. in casts.
		 */
		res->process = get_process(value->process);
		break;
	    default:
		abort();
	}
	break;
    }
    va_end(ap);
    put(value);
    return res;
}


/* ----- Operators: object access ------------------------------------------ */


struct value *op_object(const struct op *self)
{
    struct value *value;

    value = new_value(&id_type);
    value->process = get_process(__current_process());
    value->u.id.base = NULL;
    value->u.id.name = stralloc(string_eval(self->u.arg.a)); 
    return value;
}


/* ----- Operators: op_member extension ------------------------------------ */


struct value *op_object_unit(const struct op *self)
{
    struct value *value,*res;
    struct process *process;
    const char *name;

    value = evaluate(self->u.arg.a);
    process = __value_process(value);
    name = string_eval(self->u.arg.b);
    res = new_value(&location_type);
    res->process = get_process(process);
    res->u.obj = dwx_find_unit(process->objs,name);
    if (!res->u.obj) clerrorf("no such compilation unit \"%s\"",name);
    put(value);
    return res;
}


int sbug_op_member(const struct op *expr,struct value **value,const char *tag)
{
    struct value *res;

    *value = evaluate_id_left(expr);
    if ((*value)->type == &process_type) {
	res = new_value(&id_type);
	res->process = get_process(__value_process(*value));
	res->u.id.base = NULL;
    }
    else if ((*value)->type == &string_type) {
	const char *unit = __value_string(*value);
	
	res = new_value(&id_type);
	res->process = get_process(current_process());
	res->u.id.base = dwx_find_unit(res->process->objs,unit);
	if (!res->u.id.base) clerrorf("no such compilation unit \"%s\"",unit);
    }
    else if ((*value)->type == &location_type) {
	res = new_value(&id_type);
	res->process = get_process((*value)->process);
	res->u.id.base = (*value)->u.obj;
    }
    else if ((*value)->type == &id_type) {
	*value = evaluate_id(*value,dwx_obj_fn,dwx_obj_var,0);
	if ((*value)->type != &location_type) return 0;
	res = new_value(&id_type);
	res->process = get_process((*value)->process);
	res->u.id.base = (*value)->u.obj;
    }
    else return 0;
    res->u.id.name = stralloc(tag);
    put(*value);
    *value = res;
    return 1;
}
